from kio.schema.list_client_metrics_resources.v0.request import (
    ListClientMetricsResourcesRequest as ListClientMetricsResourcesRequestV0,
)
from kio.schema.list_client_metrics_resources.v0.request import (
    RequestHeader as ListClientMetricsResourcesRequestHeaderV0,
)
from kio.schema.list_client_metrics_resources.v0.response import (
    ListClientMetricsResourcesResponse as ListClientMetricsResourcesResponseV0,
)
from kio.schema.list_client_metrics_resources.v0.response import (
    ResponseHeader as ListClientMetricsResourcesResponseHeaderV0,
)

ListClientMetricsResourcesRequestHeader = ListClientMetricsResourcesRequestHeaderV0

ListClientMetricsResourcesResponseHeader = ListClientMetricsResourcesResponseHeaderV0

ListClientMetricsResourcesRequest = ListClientMetricsResourcesRequestV0

ListClientMetricsResourcesResponse = ListClientMetricsResourcesResponseV0



from kio.schema.read_share_group_state_summary.v0.request import (
    ReadShareGroupStateSummaryRequest as ReadShareGroupStateSummaryRequestV0,
)
from kio.schema.read_share_group_state_summary.v0.request import (
    RequestHeader as ReadShareGroupStateSummaryRequestHeaderV0,
)
from kio.schema.read_share_group_state_summary.v0.response import (
    ReadShareGroupStateSummaryResponse as ReadShareGroupStateSummaryResponseV0,
)
from kio.schema.read_share_group_state_summary.v0.response import (
    ResponseHeader as ReadShareGroupStateSummaryResponseHeaderV0,
)

ReadShareGroupStateSummaryRequestHeader = ReadShareGroupStateSummaryRequestHeaderV0

ReadShareGroupStateSummaryResponseHeader = ReadShareGroupStateSummaryResponseHeaderV0

ReadShareGroupStateSummaryRequest = ReadShareGroupStateSummaryRequestV0

ReadShareGroupStateSummaryResponse = ReadShareGroupStateSummaryResponseV0



from kio.schema.controlled_shutdown.v0.request import (
    ControlledShutdownRequest as ControlledShutdownRequestV0,
)
from kio.schema.controlled_shutdown.v0.request import (
    RequestHeader as ControlledShutdownRequestHeaderV0,
)
from kio.schema.controlled_shutdown.v0.response import (
    ControlledShutdownResponse as ControlledShutdownResponseV0,
)
from kio.schema.controlled_shutdown.v0.response import (
    ResponseHeader as ControlledShutdownResponseHeaderV0,
)
from kio.schema.controlled_shutdown.v1.request import (
    ControlledShutdownRequest as ControlledShutdownRequestV1,
)
from kio.schema.controlled_shutdown.v1.request import (
    RequestHeader as ControlledShutdownRequestHeaderV1,
)
from kio.schema.controlled_shutdown.v1.response import (
    ControlledShutdownResponse as ControlledShutdownResponseV1,
)
from kio.schema.controlled_shutdown.v1.response import (
    ResponseHeader as ControlledShutdownResponseHeaderV1,
)
from kio.schema.controlled_shutdown.v2.request import (
    ControlledShutdownRequest as ControlledShutdownRequestV2,
)
from kio.schema.controlled_shutdown.v2.request import (
    RequestHeader as ControlledShutdownRequestHeaderV2,
)
from kio.schema.controlled_shutdown.v2.response import (
    ControlledShutdownResponse as ControlledShutdownResponseV2,
)
from kio.schema.controlled_shutdown.v2.response import (
    ResponseHeader as ControlledShutdownResponseHeaderV2,
)
from kio.schema.controlled_shutdown.v3.request import (
    ControlledShutdownRequest as ControlledShutdownRequestV3,
)
from kio.schema.controlled_shutdown.v3.request import (
    RequestHeader as ControlledShutdownRequestHeaderV3,
)
from kio.schema.controlled_shutdown.v3.response import (
    ControlledShutdownResponse as ControlledShutdownResponseV3,
)
from kio.schema.controlled_shutdown.v3.response import (
    ResponseHeader as ControlledShutdownResponseHeaderV3,
)


ControlledShutdownRequestHeader = (
    ControlledShutdownRequestHeaderV0 | ControlledShutdownRequestHeaderV1 | ControlledShutdownRequestHeaderV2 | ControlledShutdownRequestHeaderV3
)

ControlledShutdownResponseHeader = (
    ControlledShutdownResponseHeaderV0 | ControlledShutdownResponseHeaderV1 | ControlledShutdownResponseHeaderV2 | ControlledShutdownResponseHeaderV3
)

ControlledShutdownRequest = (
    ControlledShutdownRequestV0 | ControlledShutdownRequestV1 | ControlledShutdownRequestV2 | ControlledShutdownRequestV3
)

ControlledShutdownResponse = (
    ControlledShutdownResponseV0 | ControlledShutdownResponseV1 | ControlledShutdownResponseV2 | ControlledShutdownResponseV3
)



from datetime import datetime
from typing import Optional

from sqlalchemy import (
    TIMESTAMP,
    BigInteger,
    Boolean,
    ForeignKey,
    ForeignKeyConstraint,
    Identity,
    Index,
    Integer,
    String,
    Text,
    UniqueConstraint,
    text,
    and_,
    select,
)
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship


class Base(DeclarativeBase):
    pass


class IntIdMixin:
    id: Mapped[int] = mapped_column(Integer, Identity(always=True), primary_key=True)


class BigIntIdMixin:
    id: Mapped[int] = mapped_column(
        BigInteger().with_variant(Integer, "sqlite"),
        Identity(always=True),
        primary_key=True,
    )


class TimestampMixin:
    created_at: Mapped[datetime] = mapped_column(
        TIMESTAMP(timezone=True),
        server_default=text("CURRENT_TIMESTAMP"),
        nullable=False,
    )

    updated_at: Mapped[datetime] = mapped_column(
        TIMESTAMP(timezone=True),
        server_default=text("CURRENT_TIMESTAMP"),
        onupdate=text("CURRENT_TIMESTAMP"),
        nullable=False,
    )


class Topic(Base, IntIdMixin, TimestampMixin):
    __tablename__ = "topics"

    name: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    schema: Mapped[Optional[str]] = mapped_column(Text, nullable=True)

    partitions: Mapped[list["Partition"]] = relationship(back_populates="topic")


class Partition(Base, IntIdMixin, TimestampMixin):
    __tablename__ = "partitions"

    topic_name: Mapped[str] = mapped_column(
        String,
        ForeignKey("topics.name", ondelete="CASCADE"),
        nullable=False,
    )

    partition_number: Mapped[int] = mapped_column(Integer, nullable=False)
    last_offset: Mapped[int] = mapped_column(BigInteger, nullable=False, default=-1)
    log_start_offset: Mapped[int] = mapped_column(BigInteger, nullable=False, default=0)

    topic: Mapped["Topic"] = relationship(back_populates="partitions")

    wal_file_offsets: Mapped[list["WALFileOffset"]] = relationship(
        back_populates="partition",
        primaryjoin=(
            "and_("
            "WALFileOffset.topic_name == Partition.topic_name, "
            "WALFileOffset.partition_number == Partition.partition_number"
            ")"
        ),
        foreign_keys="[WALFileOffset.topic_name, WALFileOffset.partition_number]",
    )

    __table_args__ = (
        UniqueConstraint("topic_name", "partition_number"),
        Index("ix_partition_topic_partition", "topic_name", "partition_number"),
    )


class WALFile(Base, BigIntIdMixin, TimestampMixin):
    __tablename__ = "wal_files"

    uri: Mapped[str] = mapped_column(Text, nullable=False)
    etag: Mapped[Optional[str]] = mapped_column(Text)
    total_bytes: Mapped[int] = mapped_column(BigInteger, nullable=False)
    total_messages: Mapped[int] = mapped_column(Integer, nullable=False)

    compacted_at: Mapped[Optional[datetime]] = mapped_column(TIMESTAMP(timezone=True))

    wal_file_offsets: Mapped[list["WALFileOffset"]] = relationship(
        back_populates="wal_file"
    )


class WALFileOffset(Base, BigIntIdMixin):
    __tablename__ = "wal_file_offsets"

    wal_file_id: Mapped[int] = mapped_column(
        ForeignKey("wal_files.id", ondelete="CASCADE"), nullable=False
    )

    topic_name: Mapped[str] = mapped_column(nullable=False)
    partition_number: Mapped[int] = mapped_column(nullable=False)

    base_offset: Mapped[int] = mapped_column(BigInteger, nullable=False)
    last_offset: Mapped[int] = mapped_column(BigInteger, nullable=False)

    byte_start: Mapped[int] = mapped_column(BigInteger, nullable=False)
    byte_end: Mapped[int] = mapped_column(BigInteger, nullable=False)

    min_timestamp: Mapped[Optional[int]] = mapped_column(BigInteger, nullable=True)
    max_timestamp: Mapped[Optional[int]] = mapped_column(BigInteger, nullable=True)

    wal_file: Mapped["WALFile"] = relationship(back_populates="wal_file_offsets")

    partition: Mapped["Partition"] = relationship(
        back_populates="wal_file_offsets",
        primaryjoin=and_(
            topic_name == Partition.topic_name,
            partition_number == Partition.partition_number,
        ),
        foreign_keys=[topic_name, partition_number],
    )

    __table_args__ = (
        ForeignKeyConstraint(
            ["topic_name", "partition_number"],
            ["partitions.topic_name", "partitions.partition_number"],
            ondelete="CASCADE",
        ),
        Index("ix_wal_file_offset_topic_partition", "topic_name", "partition_number"),
        Index(
            "ix_wal_file_offset_topic_part_ts",
            "topic_name",
            "partition_number",
            "min_timestamp",
            "max_timestamp",
        ),
    )


class ParquetFile(Base, BigIntIdMixin, TimestampMixin):
    __tablename__ = "parquet_files"

    topic_name: Mapped[str] = mapped_column(
        String, ForeignKey("topics.name", ondelete="CASCADE"), nullable=False
    )
    partition_number: Mapped[int] = mapped_column(Integer, nullable=False)

    uri: Mapped[str] = mapped_column(Text, nullable=False)

    total_bytes: Mapped[int] = mapped_column(BigInteger, nullable=False)
    row_count: Mapped[int] = mapped_column(BigInteger, nullable=False)

    min_offset: Mapped[int] = mapped_column(BigInteger, nullable=False)
    max_offset: Mapped[int] = mapped_column(BigInteger, nullable=False)

    min_timestamp: Mapped[datetime | None] = mapped_column(TIMESTAMP(timezone=True))
    max_timestamp: Mapped[datetime | None] = mapped_column(TIMESTAMP(timezone=True))

    generation: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    compacted_at: Mapped[datetime | None] = mapped_column(TIMESTAMP(timezone=True))

    partition: Mapped["Partition"] = relationship(
        primaryjoin=and_(
            topic_name == Partition.topic_name,
            partition_number == Partition.partition_number,
        ),
        foreign_keys=[topic_name, partition_number],
        viewonly=True,
    )

    __table_args__ = (
        ForeignKeyConstraint(
            ["topic_name", "partition_number"],
            ["partitions.topic_name", "partitions.partition_number"],
            ondelete="CASCADE",
        ),
        UniqueConstraint("uri"),
        Index("ix_pf_topic_partition", "topic_name", "partition_number"),
        Index(
            "ix_pf_topic_partition_min", "topic_name", "partition_number", "min_offset"
        ),
        Index(
            "ix_pf_topic_partition_max", "topic_name", "partition_number", "max_offset"
        ),
    )


class ParquetFileSource(Base, BigIntIdMixin, TimestampMixin):
    __tablename__ = "parquet_file_sources"
    parquet_file_id: Mapped[int] = mapped_column(
        ForeignKey("parquet_files.id", ondelete="CASCADE"), nullable=False
    )
    wal_file_id: Mapped[int] = mapped_column(
        ForeignKey("wal_files.id", ondelete="CASCADE"), nullable=False
    )
    __table_args__ = (
        UniqueConstraint("parquet_file_id", "wal_file_id"),
        Index("ix_pfs_pf", "parquet_file_id"),
        Index("ix_pfs_wf", "wal_file_id"),
    )


class ParquetFileParent(Base, BigIntIdMixin, TimestampMixin):
    __tablename__ = "parquet_file_parents"
    child_parquet_file_id: Mapped[int] = mapped_column(
        ForeignKey("parquet_files.id", ondelete="CASCADE"), nullable=False
    )
    parent_parquet_file_id: Mapped[int] = mapped_column(
        ForeignKey("parquet_files.id", ondelete="CASCADE"), nullable=False
    )
    __table_args__ = (
        UniqueConstraint("child_parquet_file_id", "parent_parquet_file_id"),
        Index("ix_pfp_child", "child_parquet_file_id"),
        Index("ix_pfp_parent", "parent_parquet_file_id"),
    )


async def assert_no_overlap(
    session, topic: str, partition: int, min_off: int, max_off: int
) -> None:
    q = (
        select(ParquetFile.id)
        .where(
            ParquetFile.topic_name == topic,
            ParquetFile.partition_number == partition,
            ParquetFile.compacted_at.is_(None),
            ParquetFile.min_offset <= max_off,
            ParquetFile.max_offset >= min_off,
        )
        .limit(1)
    )
    if (await session.execute(q)).scalar() is not None:
        raise ValueError(
            f"Overlapping Parquet range for {topic}[{partition}] {min_off}-{max_off}"
        )



from kio.schema.delete_share_group_state.v0.request import (
    DeleteShareGroupStateRequest as DeleteShareGroupStateRequestV0,
)
from kio.schema.delete_share_group_state.v0.request import (
    RequestHeader as DeleteShareGroupStateRequestHeaderV0,
)
from kio.schema.delete_share_group_state.v0.response import (
    DeleteShareGroupStateResponse as DeleteShareGroupStateResponseV0,
)
from kio.schema.delete_share_group_state.v0.response import (
    ResponseHeader as DeleteShareGroupStateResponseHeaderV0,
)

DeleteShareGroupStateRequestHeader = DeleteShareGroupStateRequestHeaderV0
DeleteShareGroupStateResponseHeader = DeleteShareGroupStateResponseHeaderV0
DeleteShareGroupStateRequest = DeleteShareGroupStateRequestV0
DeleteShareGroupStateResponse = DeleteShareGroupStateResponseV0



from kio.schema.sasl_authenticate.v0.request import (
    SaslAuthenticateRequest as SaslAuthenticateRequestV0,
)
from kio.schema.sasl_authenticate.v0.request import (
    RequestHeader as SaslAuthenticateRequestHeaderV0,
)
from kio.schema.sasl_authenticate.v0.response import (
    SaslAuthenticateResponse as SaslAuthenticateResponseV0,
)
from kio.schema.sasl_authenticate.v0.response import (
    ResponseHeader as SaslAuthenticateResponseHeaderV0,
)
from kio.schema.sasl_authenticate.v1.request import (
    SaslAuthenticateRequest as SaslAuthenticateRequestV1,
)
from kio.schema.sasl_authenticate.v1.request import (
    RequestHeader as SaslAuthenticateRequestHeaderV1,
)
from kio.schema.sasl_authenticate.v1.response import (
    SaslAuthenticateResponse as SaslAuthenticateResponseV1,
)
from kio.schema.sasl_authenticate.v1.response import (
    ResponseHeader as SaslAuthenticateResponseHeaderV1,
)
from kio.schema.sasl_authenticate.v2.request import (
    SaslAuthenticateRequest as SaslAuthenticateRequestV2,
)
from kio.schema.sasl_authenticate.v2.request import (
    RequestHeader as SaslAuthenticateRequestHeaderV2,
)
from kio.schema.sasl_authenticate.v2.response import (
    SaslAuthenticateResponse as SaslAuthenticateResponseV2,
)
from kio.schema.sasl_authenticate.v2.response import (
    ResponseHeader as SaslAuthenticateResponseHeaderV2,
)


SaslAuthenticateRequestHeader = (
    SaslAuthenticateRequestHeaderV0
    | SaslAuthenticateRequestHeaderV1
    | SaslAuthenticateRequestHeaderV2
)

SaslAuthenticateResponseHeader = (
    SaslAuthenticateResponseHeaderV0
    | SaslAuthenticateResponseHeaderV1
    | SaslAuthenticateResponseHeaderV2
)

SaslAuthenticateRequest = (
    SaslAuthenticateRequestV0 | SaslAuthenticateRequestV1 | SaslAuthenticateRequestV2
)

SaslAuthenticateResponse = (
    SaslAuthenticateResponseV0
    | SaslAuthenticateResponseV1
    | SaslAuthenticateResponseV2
)



import datetime
import io
from typing import Optional, Tuple, Sequence

from kio.schema.errors import ErrorCode
from kio.schema.list_offsets.v0.request import (
    ListOffsetsRequest as ListOffsetsRequestV0,
)
from kio.schema.list_offsets.v0.request import (
    RequestHeader as ListOffsetsRequestHeaderV0,
)
from kio.schema.list_offsets.v0.response import (
    ListOffsetsResponse as ListOffsetsResponseV0,
)
from kio.schema.list_offsets.v0.response import (
    ResponseHeader as ListOffsetsResponseHeaderV0,
)
from kio.schema.list_offsets.v1.request import (
    ListOffsetsRequest as ListOffsetsRequestV1,
)
from kio.schema.list_offsets.v1.request import (
    RequestHeader as ListOffsetsRequestHeaderV1,
)
from kio.schema.list_offsets.v1.response import (
    ListOffsetsResponse as ListOffsetsResponseV1,
)
from kio.schema.list_offsets.v1.response import (
    ResponseHeader as ListOffsetsResponseHeaderV1,
)
from kio.schema.list_offsets.v2.request import (
    ListOffsetsRequest as ListOffsetsRequestV2,
)
from kio.schema.list_offsets.v2.request import (
    RequestHeader as ListOffsetsRequestHeaderV2,
)
from kio.schema.list_offsets.v2.response import (
    ListOffsetsResponse as ListOffsetsResponseV2,
)
from kio.schema.list_offsets.v2.response import (
    ResponseHeader as ListOffsetsResponseHeaderV2,
)
from kio.schema.list_offsets.v3.request import (
    ListOffsetsRequest as ListOffsetsRequestV3,
)
from kio.schema.list_offsets.v3.request import (
    RequestHeader as ListOffsetsRequestHeaderV3,
)
from kio.schema.list_offsets.v3.response import (
    ListOffsetsResponse as ListOffsetsResponseV3,
)
from kio.schema.list_offsets.v3.response import (
    ResponseHeader as ListOffsetsResponseHeaderV3,
)
from kio.schema.list_offsets.v4.request import (
    ListOffsetsRequest as ListOffsetsRequestV4,
)
from kio.schema.list_offsets.v4.request import (
    RequestHeader as ListOffsetsRequestHeaderV4,
)
from kio.schema.list_offsets.v4.response import (
    ListOffsetsResponse as ListOffsetsResponseV4,
)
from kio.schema.list_offsets.v4.response import (
    ResponseHeader as ListOffsetsResponseHeaderV4,
)
from kio.schema.list_offsets.v5.request import (
    ListOffsetsRequest as ListOffsetsRequestV5,
)
from kio.schema.list_offsets.v5.request import (
    RequestHeader as ListOffsetsRequestHeaderV5,
)
from kio.schema.list_offsets.v5.response import (
    ListOffsetsResponse as ListOffsetsResponseV5,
)
from kio.schema.list_offsets.v5.response import (
    ResponseHeader as ListOffsetsResponseHeaderV5,
)
from kio.schema.list_offsets.v6.request import (
    ListOffsetsRequest as ListOffsetsRequestV6,
)
from kio.schema.list_offsets.v6.request import (
    RequestHeader as ListOffsetsRequestHeaderV6,
)
from kio.schema.list_offsets.v6.response import (
    ListOffsetsResponse as ListOffsetsResponseV6,
)
from kio.schema.list_offsets.v6.response import (
    ResponseHeader as ListOffsetsResponseHeaderV6,
)
from kio.schema.list_offsets.v7.request import (
    ListOffsetsRequest as ListOffsetsRequestV7,
)
from kio.schema.list_offsets.v7.request import (
    RequestHeader as ListOffsetsRequestHeaderV7,
)
from kio.schema.list_offsets.v7.response import (
    ListOffsetsResponse as ListOffsetsResponseV7,
)
from kio.schema.list_offsets.v7.response import (
    ResponseHeader as ListOffsetsResponseHeaderV7,
)
from kio.schema.list_offsets.v8.request import (
    ListOffsetsRequest as ListOffsetsRequestV8,
)
from kio.schema.list_offsets.v8.request import (
    RequestHeader as ListOffsetsRequestHeaderV8,
)
from kio.schema.list_offsets.v8.response import (
    ListOffsetsResponse as ListOffsetsResponseV8,
)
from kio.schema.list_offsets.v8.response import (
    ResponseHeader as ListOffsetsResponseHeaderV8,
)
from kio.schema.list_offsets.v9.request import (
    ListOffsetsRequest as ListOffsetsRequestV9,
)
from kio.schema.list_offsets.v9.request import (
    RequestHeader as ListOffsetsRequestHeaderV9,
)
from kio.schema.list_offsets.v9.response import (
    ListOffsetsResponse as ListOffsetsResponseV9,
)
from kio.schema.list_offsets.v9.response import (
    ResponseHeader as ListOffsetsResponseHeaderV9,
)
from kio.static.primitive import i32, i64, i32Timedelta

from icestream.config import Config
from icestream.kafkaserver.protocol import decode_kafka_records
from icestream.kafkaserver.wal.serde import decode_kafka_wal_file
from icestream.models import ParquetFile, WALFile, WALFileOffset, Partition
from icestream.utils import wal_uri_to_object_key

ListOffsetsRequestHeader = (
        ListOffsetsRequestHeaderV0
        | ListOffsetsRequestHeaderV1
        | ListOffsetsRequestHeaderV2
        | ListOffsetsRequestHeaderV3
        | ListOffsetsRequestHeaderV4
        | ListOffsetsRequestHeaderV5
        | ListOffsetsRequestHeaderV6
        | ListOffsetsRequestHeaderV7
        | ListOffsetsRequestHeaderV8
        | ListOffsetsRequestHeaderV9
)

ListOffsetsResponseHeader = (
        ListOffsetsResponseHeaderV0
        | ListOffsetsResponseHeaderV1
        | ListOffsetsResponseHeaderV2
        | ListOffsetsResponseHeaderV3
        | ListOffsetsResponseHeaderV4
        | ListOffsetsResponseHeaderV5
        | ListOffsetsResponseHeaderV6
        | ListOffsetsResponseHeaderV7
        | ListOffsetsResponseHeaderV8
        | ListOffsetsResponseHeaderV9
)

ListOffsetsRequest = (
        ListOffsetsRequestV0
        | ListOffsetsRequestV1
        | ListOffsetsRequestV2
        | ListOffsetsRequestV3
        | ListOffsetsRequestV4
        | ListOffsetsRequestV5
        | ListOffsetsRequestV6
        | ListOffsetsRequestV7
        | ListOffsetsRequestV8
        | ListOffsetsRequestV9
)

ListOffsetsResponse = (
        ListOffsetsResponseV0
        | ListOffsetsResponseV1
        | ListOffsetsResponseV2
        | ListOffsetsResponseV3
        | ListOffsetsResponseV4
        | ListOffsetsResponseV5
        | ListOffsetsResponseV6
        | ListOffsetsResponseV7
        | ListOffsetsResponseV8
        | ListOffsetsResponseV9
)

from sqlalchemy import select, asc, Result, Row

import kio.schema.list_offsets.v0 as lo_v0
import kio.schema.list_offsets.v1 as lo_v1
import kio.schema.list_offsets.v2 as lo_v2
import kio.schema.list_offsets.v3 as lo_v3
import kio.schema.list_offsets.v4 as lo_v4
import kio.schema.list_offsets.v5 as lo_v5
import kio.schema.list_offsets.v6 as lo_v6
import kio.schema.list_offsets.v7 as lo_v7
import kio.schema.list_offsets.v8 as lo_v8
import kio.schema.list_offsets.v9 as lo_v9

import pyarrow.parquet as pq

EARLIEST = -2
LATEST = -1


async def _find_offset_for_timestamp_parquet(
        config: Config, pf: ParquetFile, ts: int, floor_offset: int
) -> Optional[Tuple[int, int]]:
    if pf.min_timestamp and pf.max_timestamp:
        min_ts = int(pf.min_timestamp.timestamp() * 1000)
        max_ts = int(pf.max_timestamp.timestamp() * 1000)
        if max_ts < ts:
            return None  # skip everything older than target
        # if min_ts >= ts we still need to read to get the first offset >= floor_offset
        # (we can't assume the very first row meets floor_offset).
    obj = await config.store.get_async(wal_uri_to_object_key(config, pf.uri))
    blob = await obj.bytes_async()
    pfq = pq.ParquetFile(io.BytesIO(bytes(blob)))

    # TODO can we optimize by using arrow stats?
    cols = ["offset", "timestamp_ms"]
    for rg in range(pfq.num_row_groups):
        tbl = pfq.read_row_group(rg, columns=cols)
        off = tbl.column(tbl.schema.get_field_index("offset"))
        tscol = tbl.column(tbl.schema.get_field_index("timestamp_ms"))
        for i in range(tbl.num_rows):
            o = int(off[i].as_py())
            if o < floor_offset:
                continue
            tsv = tscol[i].as_py()
            if tsv is None:
                # if timestamp is missing, treat as 0
                # Kafka allows unknown timestamps
                tsv = 0
            tsv = int(tsv)
            if tsv >= ts:
                return o, tsv
    return None


async def _find_offset_for_timestamp_wal(
        config: Config, wf: WALFile, ts: int, floor_offset: int
) -> Optional[Tuple[int, int]]:
    obj = await config.store.get_async(wal_uri_to_object_key(config, wf.uri))
    data = await obj.bytes_async()
    decoded = decode_kafka_wal_file(bytes(data))

    for b in decoded.batches:
        krb = b.kafka_record_batch
        base = krb.base_offset
        if base is None:
            continue
        lod = krb.last_offset_delta or 0
        last_in_batch = base + lod
        if last_in_batch < floor_offset:
            continue

        base_ts = krb.base_timestamp or 0
        # decode record bodies only when batch may contain candidate records
        recs = decode_kafka_records(krb.records)

        for r in recs:
            a = base + r.offset_delta
            if a < floor_offset:
                continue
            rec_ts = base_ts + (r.timestamp_delta or 0)

            if rec_ts >= ts:
                return a, rec_ts
    return None


async def _resolve_timestamp_to_offset(
        config: Config,
        topic: str,
        partition: int,
        target_ts: int,
        last_offset: int,
        log_start: int,
) -> Tuple[ErrorCode, int, int]:
    if target_ts == LATEST:
        return ErrorCode.none, target_ts, last_offset + 1
    if target_ts == EARLIEST:
        return ErrorCode.none, target_ts, log_start

    async with config.async_session_factory() as session:
        pfiles: Sequence[ParquetFile] = (
            await session.execute(
                select(ParquetFile)
                .where(
                    ParquetFile.topic_name == topic,
                    ParquetFile.partition_number == partition,
                    ParquetFile.compacted_at.is_(None),
                    ParquetFile.max_offset >= log_start,
                    ParquetFile.min_offset <= last_offset,
                )
                .order_by(asc(ParquetFile.min_offset))
            )
        ).scalars().all()

        result: Result[Tuple[WALFileOffset, WALFile]] = await session.execute(
            select(WALFileOffset, WALFile)
            .join(WALFile, WALFile.id == WALFileOffset.wal_file_id)
            .where(
                WALFile.compacted_at.is_(None),
                WALFileOffset.topic_name == topic,
                WALFileOffset.partition_number == partition,
                WALFileOffset.last_offset >= log_start,
                WALFileOffset.base_offset <= last_offset,
            )
            .order_by(asc(WALFileOffset.base_offset))
        )
        wf_rows: Sequence[Tuple[WALFileOffset, WALFile]] = result.all()

    # track global min/max timestamps we see in any segment
    seen_min_ts: Optional[int] = None
    seen_max_ts: Optional[int] = None

    def _bump(tsv: Optional[int]):
        nonlocal seen_min_ts, seen_max_ts
        if tsv is None:
            return
        seen_min_ts = tsv if seen_min_ts is None or tsv < seen_min_ts else seen_min_ts
        seen_max_ts = tsv if seen_max_ts is None or tsv > seen_max_ts else seen_max_ts

    floor = max(log_start, 0)

    for pf in pfiles:
        min_ts_file = int(pf.min_timestamp.timestamp() * 1000) if pf.min_timestamp else None
        max_ts_file = int(pf.max_timestamp.timestamp() * 1000) if pf.max_timestamp else None
        _bump(min_ts_file)
        _bump(max_ts_file)

        if max_ts_file is not None and max_ts_file < target_ts:
            if pf.max_offset is not None:
                floor = max(floor, int(pf.max_offset) + 1)
            continue

        found = await _find_offset_for_timestamp_parquet(config, pf, target_ts, floor)
        if found is not None:
            o, tsv = found
            return ErrorCode.none, tsv, o

        if pf.max_offset is not None:
            floor = max(floor, int(pf.max_offset) + 1)

    for off_row, wf in wf_rows:
        min_ts_file = off_row.min_timestamp
        max_ts_file = off_row.max_timestamp
        _bump(min_ts_file)
        _bump(max_ts_file)

        if max_ts_file is not None and max_ts_file < target_ts:
            floor = max(floor, int(off_row.last_offset) + 1)
            continue

        found = await _find_offset_for_timestamp_wal(config, wf, target_ts, floor)
        if found is not None:
            o, tsv = found
            return ErrorCode.none, tsv, o

        floor = max(floor, int(off_row.last_offset) + 1)

    if seen_max_ts is not None and target_ts > seen_max_ts:
        return ErrorCode.none, target_ts, last_offset + 1

    if seen_min_ts is not None and target_ts <= seen_min_ts:
        return ErrorCode.none, target_ts, log_start

    return ErrorCode.none, target_ts, last_offset + 1


async def do_list_offsets(config: Config, req: ListOffsetsRequest, api_version: int) -> ListOffsetsResponse:
    topic_responses: list[lo_v9.response.ListOffsetsTopicResponse] = []
    print(req)

    async with config.async_session_factory() as session:
        for t in req.topics:
            topic_name = t.name
            part_responses: list[lo_v9.response.ListOffsetsPartitionResponse] = []

            for pr in t.partitions:
                partition = int(pr.partition_index)
                # request timestamp (ms or sentinel)
                ts = int(pr.timestamp)

                row = (
                    await session.execute(
                        select(Partition.last_offset, Partition.log_start_offset)
                        .where(
                            Partition.topic_name == topic_name,
                            Partition.partition_number == partition,
                        )
                    )
                ).one_or_none()

                if row is None:
                    part_responses.append(
                        lo_v9.response.ListOffsetsPartitionResponse(
                            partition_index=i32(partition),
                            error_code=ErrorCode.unknown_topic_or_partition,
                            timestamp=i64(-1),
                            offset=i64(-1),
                        )
                    )
                    continue

                last_offset, log_start = int(row[0]), int(row[1])

                err, ret_ts, offset = await _resolve_timestamp_to_offset(
                    config=config,
                    topic=topic_name,
                    partition=partition,
                    target_ts=ts,
                    last_offset=last_offset,
                    log_start=log_start,
                )

                part_responses.append(
                    lo_v9.response.ListOffsetsPartitionResponse(
                        partition_index=i32(partition),
                        error_code=err,
                        timestamp=i64(ret_ts),
                        offset=i64(offset),
                    )
                )

            topic_responses.append(
                lo_v9.response.ListOffsetsTopicResponse(
                    name=topic_name,
                    partitions=tuple(part_responses),
                )
            )

    resp_v9 = lo_v9.ListOffsetsResponse(
        throttle_time=i32Timedelta.parse(datetime.timedelta(milliseconds=0)),
        topics=tuple(topic_responses),
    )

    if api_version == 9:
        return resp_v9
    else:
        return _list_offsets_response_ladder(resp_v9, api_version)


def _list_offsets_response_ladder(resp_v9: lo_v9.ListOffsetsResponse, api_version: int) -> ListOffsetsResponse:
    """
    Down-convert the canonical v9 response to earlier versions.
    Shapes between versions vary slightly in Kafka; the following mapping keeps the same
    semantics (per-partition {error_code, timestamp, offset}) while adapting field names.
    Adjust inner class names if your generated schema differs.
    """
    if api_version == 8:
        topics = tuple(
            lo_v8.response.ListOffsetsTopicResponse(
                name=t.name,
                partitions=tuple(
                    lo_v8.response.ListOffsetsPartitionResponse(
                        partition_index=p.partition_index,
                        error_code=p.error_code,
                        timestamp=p.timestamp,
                        offset=p.offset,
                    ) for p in t.partitions
                )
            ) for t in resp_v9.topics
        )
        return lo_v8.ListOffsetsResponse(
            throttle_time=resp_v9.throttle_time,
            topics=topics,
        )

    if api_version == 7:
        topics = tuple(
            lo_v7.response.ListOffsetsTopicResponse(
                name=t.name,
                partitions=tuple(
                    lo_v7.response.ListOffsetsPartitionResponse(
                        partition_index=p.partition_index,
                        error_code=p.error_code,
                        timestamp=p.timestamp,
                        offset=p.offset,
                    ) for p in t.partitions
                )
            ) for t in resp_v9.topics
        )
        return lo_v7.ListOffsetsResponse(
            throttle_time=resp_v9.throttle_time,
            topics=topics,
        )

    if api_version == 6:
        topics = tuple(
            lo_v6.response.ListOffsetsTopicResponse(
                name=t.name,
                partitions=tuple(
                    lo_v6.response.ListOffsetsPartitionResponse(
                        partition_index=p.partition_index,
                        error_code=p.error_code,
                        timestamp=p.timestamp,
                        offset=p.offset,
                    ) for p in t.partitions
                )
            ) for t in resp_v9.topics
        )
        return lo_v6.ListOffsetsResponse(
            throttle_time=resp_v9.throttle_time,
            topics=topics,
        )

    if api_version == 5:
        topics = tuple(
            lo_v5.response.ListOffsetsTopicResponse(
                name=t.name,
                partitions=tuple(
                    lo_v5.response.ListOffsetsPartitionResponse(
                        partition_index=p.partition_index,
                        error_code=p.error_code,
                        timestamp=p.timestamp,
                        offset=p.offset,
                    ) for p in t.partitions
                )
            ) for t in resp_v9.topics
        )
        return lo_v5.ListOffsetsResponse(
            throttle_time=resp_v9.throttle_time,
            topics=topics,
        )

    if api_version == 4:
        topics = tuple(
            lo_v4.response.ListOffsetsTopicResponse(
                name=t.name,
                partitions=tuple(
                    lo_v4.response.ListOffsetsPartitionResponse(
                        partition_index=p.partition_index,
                        error_code=p.error_code,
                        timestamp=p.timestamp,
                        offset=p.offset,
                    ) for p in t.partitions
                )
            ) for t in resp_v9.topics
        )
        return lo_v4.ListOffsetsResponse(
            throttle_time=resp_v9.throttle_time,
            topics=topics,
        )

    if api_version == 3:
        topics = tuple(
            lo_v3.response.ListOffsetsTopicResponse(
                name=t.name,
                partitions=tuple(
                    lo_v3.response.ListOffsetsPartitionResponse(
                        partition_index=p.partition_index,
                        error_code=p.error_code,
                        timestamp=p.timestamp,
                        offset=p.offset,
                    ) for p in t.partitions
                )
            ) for t in resp_v9.topics
        )
        return lo_v3.ListOffsetsResponse(
            throttle_time=resp_v9.throttle_time,
            topics=topics,
        )

    if api_version == 2:
        topics = tuple(
            lo_v2.response.ListOffsetsTopicResponse(
                name=t.name,
                partitions=tuple(
                    lo_v2.response.ListOffsetsPartitionResponse(
                        partition_index=p.partition_index,
                        error_code=p.error_code,
                        timestamp=p.timestamp,
                        offset=p.offset,
                    ) for p in t.partitions
                )
            ) for t in resp_v9.topics
        )
        return lo_v2.ListOffsetsResponse(
            throttle_time=resp_v9.throttle_time,
            topics=topics,
        )

    # v0/v1 historically had arrays of offsets; here we adapt by returning a single-element vector
    if api_version == 1:
        topics = tuple(
            lo_v1.response.ListOffsetsTopicResponse(
                name=t.name,
                partitions=tuple(
                    lo_v1.response.ListOffsetsPartitionResponse(
                        partition_index=p.partition_index,
                        error_code=p.error_code,
                        timestamp=p.timestamp,
                        offset=p.offset,  # single element, Kafka brokers also often return 1
                    ) for p in t.partitions
                )
            ) for t in resp_v9.topics
        )
        return lo_v1.ListOffsetsResponse(
            topics=topics,
        )

    if api_version == 0:
        topics = tuple(
            lo_v0.response.ListOffsetsTopicResponse(
                name=t.name,
                partitions=tuple(
                    lo_v0.response.ListOffsetsPartitionResponse(
                        partition_index=p.partition_index,
                        error_code=p.error_code,
                        old_style_offsets=(p.offset,),
                    ) for p in t.partitions
                )
            ) for t in resp_v9.topics
        )
        return lo_v0.ListOffsetsResponse(
            topics=topics,
        )

    raise ValueError(f"unsupported list_offsets api version: {api_version}")



from kio.schema.share_group_heartbeat.v0.request import (
    ShareGroupHeartbeatRequest as ShareGroupHeartbeatRequestV0,
)
from kio.schema.share_group_heartbeat.v0.request import (
    RequestHeader as ShareGroupHeartbeatRequestHeaderV0,
)
from kio.schema.share_group_heartbeat.v0.response import (
    ShareGroupHeartbeatResponse as ShareGroupHeartbeatResponseV0,
)
from kio.schema.share_group_heartbeat.v0.response import (
    ResponseHeader as ShareGroupHeartbeatResponseHeaderV0,
)


ShareGroupHeartbeatRequestHeader = ShareGroupHeartbeatRequestHeaderV0

ShareGroupHeartbeatResponseHeader = ShareGroupHeartbeatResponseHeaderV0

ShareGroupHeartbeatRequest = ShareGroupHeartbeatRequestV0

ShareGroupHeartbeatResponse = ShareGroupHeartbeatResponseV0



"""add min/max timestamps

Revision ID: 9f0f8ebe3914
Revises: 298875b23096
Create Date: 2025-10-03 14:36:51.645541

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '9f0f8ebe3914'
down_revision: Union[str, None] = '298875b23096'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('parquet_files',
    sa.Column('topic_name', sa.String(), nullable=False),
    sa.Column('partition_number', sa.Integer(), nullable=False),
    sa.Column('uri', sa.Text(), nullable=False),
    sa.Column('total_bytes', sa.BigInteger(), nullable=False),
    sa.Column('row_count', sa.BigInteger(), nullable=False),
    sa.Column('min_offset', sa.BigInteger(), nullable=False),
    sa.Column('max_offset', sa.BigInteger(), nullable=False),
    sa.Column('min_timestamp', sa.TIMESTAMP(timezone=True), nullable=True),
    sa.Column('max_timestamp', sa.TIMESTAMP(timezone=True), nullable=True),
    sa.Column('generation', sa.Integer(), nullable=False),
    sa.Column('compacted_at', sa.TIMESTAMP(timezone=True), nullable=True),
    sa.Column('id', sa.BigInteger().with_variant(sa.Integer(), 'sqlite'), sa.Identity(always=True), nullable=False),
    sa.Column('created_at', sa.TIMESTAMP(timezone=True), server_default=sa.text('CURRENT_TIMESTAMP'), nullable=False),
    sa.Column('updated_at', sa.TIMESTAMP(timezone=True), server_default=sa.text('CURRENT_TIMESTAMP'), nullable=False),
    sa.ForeignKeyConstraint(['topic_name', 'partition_number'], ['partitions.topic_name', 'partitions.partition_number'], ondelete='CASCADE'),
    sa.ForeignKeyConstraint(['topic_name'], ['topics.name'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('uri')
    )
    op.create_index('ix_pf_topic_partition', 'parquet_files', ['topic_name', 'partition_number'], unique=False)
    op.create_index('ix_pf_topic_partition_max', 'parquet_files', ['topic_name', 'partition_number', 'max_offset'], unique=False)
    op.create_index('ix_pf_topic_partition_min', 'parquet_files', ['topic_name', 'partition_number', 'min_offset'], unique=False)
    op.create_table('parquet_file_parents',
    sa.Column('child_parquet_file_id', sa.BigInteger().with_variant(sa.Integer(), 'sqlite'), nullable=False),
    sa.Column('parent_parquet_file_id', sa.BigInteger().with_variant(sa.Integer(), 'sqlite'), nullable=False),
    sa.Column('id', sa.BigInteger().with_variant(sa.Integer(), 'sqlite'), sa.Identity(always=True), nullable=False),
    sa.Column('created_at', sa.TIMESTAMP(timezone=True), server_default=sa.text('CURRENT_TIMESTAMP'), nullable=False),
    sa.Column('updated_at', sa.TIMESTAMP(timezone=True), server_default=sa.text('CURRENT_TIMESTAMP'), nullable=False),
    sa.ForeignKeyConstraint(['child_parquet_file_id'], ['parquet_files.id'], ondelete='CASCADE'),
    sa.ForeignKeyConstraint(['parent_parquet_file_id'], ['parquet_files.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('child_parquet_file_id', 'parent_parquet_file_id')
    )
    op.create_index('ix_pfp_child', 'parquet_file_parents', ['child_parquet_file_id'], unique=False)
    op.create_index('ix_pfp_parent', 'parquet_file_parents', ['parent_parquet_file_id'], unique=False)
    op.create_table('parquet_file_sources',
    sa.Column('parquet_file_id', sa.BigInteger().with_variant(sa.Integer(), 'sqlite'), nullable=False),
    sa.Column('wal_file_id', sa.BigInteger().with_variant(sa.Integer(), 'sqlite'), nullable=False),
    sa.Column('id', sa.BigInteger().with_variant(sa.Integer(), 'sqlite'), sa.Identity(always=True), nullable=False),
    sa.Column('created_at', sa.TIMESTAMP(timezone=True), server_default=sa.text('CURRENT_TIMESTAMP'), nullable=False),
    sa.Column('updated_at', sa.TIMESTAMP(timezone=True), server_default=sa.text('CURRENT_TIMESTAMP'), nullable=False),
    sa.ForeignKeyConstraint(['parquet_file_id'], ['parquet_files.id'], ondelete='CASCADE'),
    sa.ForeignKeyConstraint(['wal_file_id'], ['wal_files.id'], ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('parquet_file_id', 'wal_file_id')
    )
    op.create_index('ix_pfs_pf', 'parquet_file_sources', ['parquet_file_id'], unique=False)
    op.create_index('ix_pfs_wf', 'parquet_file_sources', ['wal_file_id'], unique=False)
    op.add_column('topics', sa.Column('schema', sa.Text(), nullable=True))
    op.add_column('wal_file_offsets', sa.Column('min_timestamp', sa.BigInteger(), nullable=True))
    op.add_column('wal_file_offsets', sa.Column('max_timestamp', sa.BigInteger(), nullable=True))
    op.create_index('ix_wal_file_offset_topic_part_ts', 'wal_file_offsets', ['topic_name', 'partition_number', 'min_timestamp', 'max_timestamp'], unique=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('ix_wal_file_offset_topic_part_ts', table_name='wal_file_offsets')
    op.drop_column('wal_file_offsets', 'max_timestamp')
    op.drop_column('wal_file_offsets', 'min_timestamp')
    op.drop_column('topics', 'schema')
    op.drop_index('ix_pfs_wf', table_name='parquet_file_sources')
    op.drop_index('ix_pfs_pf', table_name='parquet_file_sources')
    op.drop_table('parquet_file_sources')
    op.drop_index('ix_pfp_parent', table_name='parquet_file_parents')
    op.drop_index('ix_pfp_child', table_name='parquet_file_parents')
    op.drop_table('parquet_file_parents')
    op.drop_index('ix_pf_topic_partition_min', table_name='parquet_files')
    op.drop_index('ix_pf_topic_partition_max', table_name='parquet_files')
    op.drop_index('ix_pf_topic_partition', table_name='parquet_files')
    op.drop_table('parquet_files')
    # ### end Alembic commands ###



from kio.schema.alter_partition.v0.request import (
    AlterPartitionRequest as AlterPartitionRequestV0,
)
from kio.schema.alter_partition.v0.request import (
    RequestHeader as AlterPartitionRequestHeaderV0,
)
from kio.schema.alter_partition.v0.response import (
    AlterPartitionResponse as AlterPartitionResponseV0,
)
from kio.schema.alter_partition.v0.response import (
    ResponseHeader as AlterPartitionResponseHeaderV0,
)
from kio.schema.alter_partition.v1.request import (
    AlterPartitionRequest as AlterPartitionRequestV1,
)
from kio.schema.alter_partition.v1.request import (
    RequestHeader as AlterPartitionRequestHeaderV1,
)
from kio.schema.alter_partition.v1.response import (
    AlterPartitionResponse as AlterPartitionResponseV1,
)
from kio.schema.alter_partition.v1.response import (
    ResponseHeader as AlterPartitionResponseHeaderV1,
)
from kio.schema.alter_partition.v2.request import (
    AlterPartitionRequest as AlterPartitionRequestV2,
)
from kio.schema.alter_partition.v2.request import (
    RequestHeader as AlterPartitionRequestHeaderV2,
)
from kio.schema.alter_partition.v2.response import (
    AlterPartitionResponse as AlterPartitionResponseV2,
)
from kio.schema.alter_partition.v2.response import (
    ResponseHeader as AlterPartitionResponseHeaderV2,
)
from kio.schema.alter_partition.v3.request import (
    AlterPartitionRequest as AlterPartitionRequestV3,
)
from kio.schema.alter_partition.v3.request import (
    RequestHeader as AlterPartitionRequestHeaderV3,
)
from kio.schema.alter_partition.v3.response import (
    AlterPartitionResponse as AlterPartitionResponseV3,
)
from kio.schema.alter_partition.v3.response import (
    ResponseHeader as AlterPartitionResponseHeaderV3,
)

AlterPartitionRequestHeader = (
    AlterPartitionRequestHeaderV0 | AlterPartitionRequestHeaderV1 | AlterPartitionRequestHeaderV2 | AlterPartitionRequestHeaderV3
)

AlterPartitionResponseHeader = (
    AlterPartitionResponseHeaderV0 | AlterPartitionResponseHeaderV1 | AlterPartitionResponseHeaderV2 | AlterPartitionResponseHeaderV3
)

AlterPartitionRequest = (
    AlterPartitionRequestV0 | AlterPartitionRequestV1 | AlterPartitionRequestV2 | AlterPartitionRequestV3
)

AlterPartitionResponse = (
    AlterPartitionResponseV0 | AlterPartitionResponseV1 | AlterPartitionResponseV2 | AlterPartitionResponseV3
)



from kio.schema.describe_topic_partitions.v0.request import (
    DescribeTopicPartitionsRequest as DescribeTopicPartitionsRequestV0,
)
from kio.schema.describe_topic_partitions.v0.request import (
    RequestHeader as DescribeTopicPartitionsRequestHeaderV0,
)
from kio.schema.describe_topic_partitions.v0.response import (
    DescribeTopicPartitionsResponse as DescribeTopicPartitionsResponseV0,
)
from kio.schema.describe_topic_partitions.v0.response import (
    ResponseHeader as DescribeTopicPartitionsResponseHeaderV0,
)

DescribeTopicPartitionsRequestHeader = DescribeTopicPartitionsRequestHeaderV0
DescribeTopicPartitionsResponseHeader = DescribeTopicPartitionsResponseHeaderV0
DescribeTopicPartitionsRequest = DescribeTopicPartitionsRequestV0
DescribeTopicPartitionsResponse = DescribeTopicPartitionsResponseV0



import datetime
import io
from collections import defaultdict
from typing import Dict, List, Tuple

import pyarrow as pa
import pyarrow.parquet as pq

from icestream.compaction import build_uri
from icestream.compaction.types import CompactionContext, CompactionProcessor
from icestream.compaction.schema import PARQUET_RECORD_SCHEMA
from icestream.kafkaserver.protocol import decode_kafka_records
from icestream.models import ParquetFile, ParquetFileSource, assert_no_overlap


class WalToParquetProcessor(CompactionProcessor):
    async def apply(self, ctx: CompactionContext) -> None:
        if not ctx.wal_decoded:
            return

        buckets = self.bucket_records(ctx)

        for (topic, partition), rows in buckets.items():
            if not rows:
                continue

            rows.sort(key=lambda r: r["offset"])
            await self._flush_buckets(ctx, topic, partition, rows)

    @staticmethod
    def _estimate_record_size(r: dict) -> int:
        k = len(r["key"]) if r["key"] else 0
        v = len(r["value"]) if r["value"] else 0
        h = sum(
            (len(hd["key"]) + (len(hd["value"]) if hd["value"] else 0))
            for hd in (r["headers"] or [])
        )
        return k + v + h + 64  # include small buffer

    @staticmethod
    def bucket_records(ctx: CompactionContext) -> Dict[Tuple[str, int], List[dict]]:
        buckets: Dict[Tuple[str, int], List[dict]] = defaultdict(list)

        for wf in ctx.wal_decoded:
            for batch in wf.batches:
                topic = batch.topic
                partition = batch.partition
                base = batch.kafka_record_batch.base_offset
                if base is None:
                    continue

                records = decode_kafka_records(batch.kafka_record_batch.records)
                for rec in records:
                    offset = base + rec.offset_delta
                    ts_ms = batch.kafka_record_batch.base_timestamp
                    if ts_ms is not None:
                        ts_ms = ts_ms + rec.timestamp_delta

                    buckets[(topic, partition)].append(
                        {
                            "partition": partition,
                            "offset": offset,
                            "timestamp_ms": ts_ms,
                            "key": rec.key,
                            "value": rec.value,
                            "headers": [
                                           {"key": h.key, "value": h.value} for h in rec.headers
                                       ]
                                       or None,
                        }
                    )

        return buckets

    async def _flush_buckets(
            self, ctx: CompactionContext, topic: str, partition: int, rows: List[dict]
    ):
        target_bytes = ctx.config.PARQUET_TARGET_FILE_BYTES

        chunk: list[dict] = []
        approx = 0

        for r in rows:
            approx += self._estimate_record_size(r)
            chunk.append(r)
            if approx >= target_bytes:
                await self._flush_chunk(ctx, topic, partition, chunk)
                chunk, approx = [], 0

        if chunk:
            await self._flush_chunk(ctx, topic, partition, chunk)

    async def _flush_chunk(
            self, ctx: CompactionContext, topic: str, partition: int, chunk_rows: List[dict]
    ):
        if not chunk_rows:
            return

        # noinspection PyArgumentList
        table = pa.Table.from_pylist(chunk_rows, schema=PARQUET_RECORD_SCHEMA)

        # row group size estimation
        sample = chunk_rows[: min(1000, len(chunk_rows))]
        kvh = sum(self._estimate_record_size(r) for r in sample)
        avg_row = max(64, kvh // max(1, len(sample)))
        rows_per_rg = max(1, ctx.config.PARQUET_ROW_GROUP_TARGET_BYTES // avg_row)

        sink = io.BytesIO()
        pq.write_table(
            table,
            sink,
            compression="zstd",
            use_dictionary=True,
            write_statistics=True,
            row_group_size=rows_per_rg,
        )
        data = sink.getvalue()
        total_bytes = len(data)
        min_off = chunk_rows[0]["offset"]
        max_off = chunk_rows[-1]["offset"]

        ts_vals = [
            r["timestamp_ms"] for r in chunk_rows if r["timestamp_ms"] is not None
        ]
        min_ts = (
            datetime.datetime.fromtimestamp(min(ts_vals) / 1000, tz=datetime.UTC)
            if ts_vals
            else None
        )
        max_ts = (
            datetime.datetime.fromtimestamp(max(ts_vals) / 1000, tz=datetime.UTC)
            if ts_vals
            else None
        )

        key = (
                ctx.config.PARQUET_PREFIX.rstrip("/")
                + f"/topics/{topic}/partition={partition}/{min_off}-{max_off}-gen0.parquet"
        )
        await ctx.config.store.put_async(key, io.BytesIO(data))

        async with ctx.config.async_session_factory() as session:
            await assert_no_overlap(session, topic, partition, min_off, max_off)
            pf = ParquetFile(
                topic_name=topic,
                partition_number=partition,
                uri=build_uri(ctx.config, key),
                total_bytes=total_bytes,
                row_count=table.num_rows,
                min_offset=min_off,
                max_offset=max_off,
                min_timestamp=min_ts,
                max_timestamp=max_ts,
                generation=0,
            )
            session.add(pf)
            await session.flush()

            for wm in ctx.wal_models:
                session.add(ParquetFileSource(parquet_file_id=pf.id, wal_file_id=wm.id))



from kio.schema.push_telemetry.v0.request import (
    PushTelemetryRequest as PushTelemetryRequestV0,
)
from kio.schema.push_telemetry.v0.request import (
    RequestHeader as PushTelemetryRequestHeaderV0,
)
from kio.schema.push_telemetry.v0.response import (
    PushTelemetryResponse as PushTelemetryResponseV0,
)
from kio.schema.push_telemetry.v0.response import (
    ResponseHeader as PushTelemetryResponseHeaderV0,
)

PushTelemetryRequestHeader = PushTelemetryRequestHeaderV0

PushTelemetryResponseHeader = PushTelemetryResponseHeaderV0

PushTelemetryRequest = PushTelemetryRequestV0

PushTelemetryResponse = PushTelemetryResponseV0



from kio.schema.alter_replica_log_dirs.v0.request import (
    AlterReplicaLogDirsRequest as AlterReplicaLogDirsRequestV0,
)
from kio.schema.alter_replica_log_dirs.v0.request import (
    RequestHeader as AlterReplicaLogDirsRequestHeaderV0,
)
from kio.schema.alter_replica_log_dirs.v0.response import (
    AlterReplicaLogDirsResponse as AlterReplicaLogDirsResponseV0,
)
from kio.schema.alter_replica_log_dirs.v0.response import (
    ResponseHeader as AlterReplicaLogDirsResponseHeaderV0,
)
from kio.schema.alter_replica_log_dirs.v1.request import (
    AlterReplicaLogDirsRequest as AlterReplicaLogDirsRequestV1,
)
from kio.schema.alter_replica_log_dirs.v1.request import (
    RequestHeader as AlterReplicaLogDirsRequestHeaderV1,
)
from kio.schema.alter_replica_log_dirs.v1.response import (
    AlterReplicaLogDirsResponse as AlterReplicaLogDirsResponseV1,
)
from kio.schema.alter_replica_log_dirs.v1.response import (
    ResponseHeader as AlterReplicaLogDirsResponseHeaderV1,
)
from kio.schema.alter_replica_log_dirs.v2.request import (
    AlterReplicaLogDirsRequest as AlterReplicaLogDirsRequestV2,
)
from kio.schema.alter_replica_log_dirs.v2.request import (
    RequestHeader as AlterReplicaLogDirsRequestHeaderV2,
)
from kio.schema.alter_replica_log_dirs.v2.response import (
    AlterReplicaLogDirsResponse as AlterReplicaLogDirsResponseV2,
)
from kio.schema.alter_replica_log_dirs.v2.response import (
    ResponseHeader as AlterReplicaLogDirsResponseHeaderV2,
)


AlterReplicaLogDirsRequestHeader = (
    AlterReplicaLogDirsRequestHeaderV0 | AlterReplicaLogDirsRequestHeaderV1 | AlterReplicaLogDirsRequestHeaderV2
)

AlterReplicaLogDirsResponseHeader = (
    AlterReplicaLogDirsResponseHeaderV0 | AlterReplicaLogDirsResponseHeaderV1 | AlterReplicaLogDirsResponseHeaderV2
)

AlterReplicaLogDirsRequest = (
    AlterReplicaLogDirsRequestV0 | AlterReplicaLogDirsRequestV1 | AlterReplicaLogDirsRequestV2
)

AlterReplicaLogDirsResponse = (
    AlterReplicaLogDirsResponseV0 | AlterReplicaLogDirsResponseV1 | AlterReplicaLogDirsResponseV2
)



from kio.schema.sasl_handshake.v0.request import (
    SaslHandshakeRequest as SaslHandshakeRequestV0,
)
from kio.schema.sasl_handshake.v0.request import (
    RequestHeader as SaslHandshakeRequestHeaderV0,
)
from kio.schema.sasl_handshake.v0.response import (
    SaslHandshakeResponse as SaslHandshakeResponseV0,
)
from kio.schema.sasl_handshake.v0.response import (
    ResponseHeader as SaslHandshakeResponseHeaderV0,
)
from kio.schema.sasl_handshake.v1.request import (
    SaslHandshakeRequest as SaslHandshakeRequestV1,
)
from kio.schema.sasl_handshake.v1.request import (
    RequestHeader as SaslHandshakeRequestHeaderV1,
)
from kio.schema.sasl_handshake.v1.response import (
    SaslHandshakeResponse as SaslHandshakeResponseV1,
)
from kio.schema.sasl_handshake.v1.response import (
    ResponseHeader as SaslHandshakeResponseHeaderV1,
)


SaslHandshakeRequestHeader = (
    SaslHandshakeRequestHeaderV0 | SaslHandshakeRequestHeaderV1
)

SaslHandshakeResponseHeader = (
    SaslHandshakeResponseHeaderV0 | SaslHandshakeResponseHeaderV1
)

SaslHandshakeRequest = SaslHandshakeRequestV0 | SaslHandshakeRequestV1

SaslHandshakeResponse = SaslHandshakeResponseV0 | SaslHandshakeResponseV1



import os

from alembic import command
from alembic.config import Config
from sqlalchemy.engine.url import make_url

from icestream.config import Config as IcestreamConfig
from icestream.models import Base


async def run_migrations(icestream_config: IcestreamConfig):
    url = make_url(icestream_config.DATABASE_URL)

    if url.drivername.startswith("sqlite"):
        async with icestream_config.engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

        return

    base_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "../../../"))
    alembic_ini_path = os.path.join(base_dir, "alembic.ini")

    alembic_cfg = Config(alembic_ini_path)
    script_location = os.path.join(base_dir, "src", "icestream", "alembic")
    alembic_cfg.set_main_option("script_location", script_location)

    command.upgrade(alembic_cfg, "head")



import asyncio
from dataclasses import dataclass

from icestream.kafkaserver.protocol import KafkaRecordBatch


@dataclass
class ProduceTopicPartitionData:
    topic: str
    partition: int
    kafka_record_batch: KafkaRecordBatch
    flush_result: asyncio.Future[bool]



from kio.schema.describe_acls.v0.request import (
    DescribeAclsRequest as DescribeAclsRequestV0,
)
from kio.schema.describe_acls.v0.request import (
    RequestHeader as DescribeAclsRequestHeaderV0,
)
from kio.schema.describe_acls.v0.response import (
    DescribeAclsResponse as DescribeAclsResponseV0,
)
from kio.schema.describe_acls.v0.response import (
    ResponseHeader as DescribeAclsResponseHeaderV0,
)
from kio.schema.describe_acls.v1.request import (
    DescribeAclsRequest as DescribeAclsRequestV1,
)
from kio.schema.describe_acls.v1.request import (
    RequestHeader as DescribeAclsRequestHeaderV1,
)
from kio.schema.describe_acls.v1.response import (
    DescribeAclsResponse as DescribeAclsResponseV1,
)
from kio.schema.describe_acls.v1.response import (
    ResponseHeader as DescribeAclsResponseHeaderV1,
)
from kio.schema.describe_acls.v2.request import (
    DescribeAclsRequest as DescribeAclsRequestV2,
)
from kio.schema.describe_acls.v2.request import (
    RequestHeader as DescribeAclsRequestHeaderV2,
)
from kio.schema.describe_acls.v2.response import (
    DescribeAclsResponse as DescribeAclsResponseV2,
)
from kio.schema.describe_acls.v2.response import (
    ResponseHeader as DescribeAclsResponseHeaderV2,
)
from kio.schema.describe_acls.v3.request import (
    DescribeAclsRequest as DescribeAclsRequestV3,
)
from kio.schema.describe_acls.v3.request import (
    RequestHeader as DescribeAclsRequestHeaderV3,
)
from kio.schema.describe_acls.v3.response import (
    DescribeAclsResponse as DescribeAclsResponseV3,
)
from kio.schema.describe_acls.v3.response import (
    ResponseHeader as DescribeAclsResponseHeaderV3,
)

DescribeAclsRequestHeader = (
    DescribeAclsRequestHeaderV0
    | DescribeAclsRequestHeaderV1
    | DescribeAclsRequestHeaderV2
    | DescribeAclsRequestHeaderV3
)

DescribeAclsResponseHeader = (
    DescribeAclsResponseHeaderV0
    | DescribeAclsResponseHeaderV1
    | DescribeAclsResponseHeaderV2
    | DescribeAclsResponseHeaderV3
)

DescribeAclsRequest = (
    DescribeAclsRequestV0
    | DescribeAclsRequestV1
    | DescribeAclsRequestV2
    | DescribeAclsRequestV3
)

DescribeAclsResponse = (
    DescribeAclsResponseV0
    | DescribeAclsResponseV1
    | DescribeAclsResponseV2
    | DescribeAclsResponseV3
)



import asyncio
import datetime
import time
from collections import defaultdict
from typing import Sequence

import httpx
from sqlalchemy import select
from sqlalchemy.orm import joinedload, selectinload

from icestream.compaction.types import CompactionProcessor, CompactionContext
from icestream.config import Config
from icestream.kafkaserver.wal import WALFile as DecodedWALFile, WALBatch
from icestream.kafkaserver.wal.serde import decode_kafka_wal_file
from icestream.models import WALFile as WALFileModel, Topic, ParquetFile
from icestream.logger import log
from icestream.utils import wal_uri_to_object_key


class CompactorWorker:
    def __init__(
            self,
            config: Config,
            processors: list[CompactionProcessor],
            time_source=time.monotonic,
    ):
        self.config = config
        self.time_source = time_source
        self.processors = processors

    async def run(self):
        log.info("CompactorWorker started")
        try:
            while True:
                await self.run_once()
        except asyncio.CancelledError:
            pass
        finally:
            log.info("CompactorWorker stopped")

    async def run_once(self, now: float | None = None):
        start_time = now or self.time_source()

        async with self.config.async_session_factory() as session:
            wal_models, total_bytes = await self._select_wal_models(session)

            if not wal_models:
                await self._sleep_until_next_interval(start_time)
                return

            wal_decoded: list[DecodedWALFile] = await self._fetch_and_decode(wal_models)
            parquet_candidates = await self._select_parquet_candidates(session)

            ctx = CompactionContext(
                config=self.config,
                wal_models=wal_models,
                wal_decoded=wal_decoded,
                parquet_candidates=parquet_candidates,
                now_monotonic=start_time,
            )

            for p in self.processors:
                await p.apply(ctx)

            now_ts = datetime.datetime.now(datetime.UTC)
            for wm in wal_models:
                if wm.compacted_at is None:
                    wm.compacted_at = now_ts
            await session.commit()

        await self._sleep_until_next_interval(start_time)

    async def _sleep_until_next_interval(self, start_time: float):
        elapsed = self.time_source() - start_time
        sleep_time = max(0, self.config.COMPACTION_INTERVAL - elapsed)
        await asyncio.sleep(sleep_time)

    async def _select_wal_models(self, session) -> tuple[Sequence[WALFileModel], int]:
        total_bytes = 0
        selected: list[WALFileModel] = []
        seen_ids: set[int] = set()

        while (
                total_bytes < self.config.MAX_COMPACTION_BYTES
                and len(selected) < self.config.MAX_COMPACTION_WAL_FILES
        ):
            limit = min(
                self.config.MAX_COMPACTION_SELECT_LIMIT,
                self.config.MAX_COMPACTION_WAL_FILES - len(selected),
            )

            id_rows = await session.scalars(
                select(WALFileModel.id)
                .where(
                    WALFileModel.compacted_at.is_(None),
                    ~WALFileModel.id.in_(seen_ids)
                )
                .order_by(WALFileModel.id)
                .with_for_update(skip_locked=True)
                .limit(limit)
            )
            ids = list(id_rows)
            if not ids:
                break
            seen_ids.update(ids)

            models = (
                await session.execute(
                    select(WALFileModel)
                    .options(selectinload(WALFileModel.wal_file_offsets))
                    .where(WALFileModel.id.in_(ids))
                    .order_by(WALFileModel.id)
                )
            ).scalars().all()

            for wal in models:
                if len(selected) >= self.config.MAX_COMPACTION_WAL_FILES:
                    break
                if total_bytes + wal.total_bytes > self.config.MAX_COMPACTION_BYTES:
                    continue
                selected.append(wal)
                total_bytes += wal.total_bytes

        return selected, total_bytes

    async def _fetch_and_decode(
            self, wal_models: Sequence[WALFileModel]
    ) -> list[DecodedWALFile]:
        decoded: list[DecodedWALFile] = []
        for wal in wal_models:
            object_key = self._wal_uri_to_object_key(wal.uri)
            get_result = await self.config.store.get_async(object_key)
            data = await get_result.bytes_async()
            decoded_file = decode_kafka_wal_file(bytes(data))
            decoded_file.id = wal.id  # type: ignore[attr-defined]
            decoded.append(decoded_file)
        return decoded

    async def _select_parquet_candidates(self, session):
        rows = (
            (
                await session.execute(
                    select(ParquetFile)
                    .where(ParquetFile.compacted_at.is_(None))
                    .order_by(
                        ParquetFile.topic_name,
                        ParquetFile.partition_number,
                        ParquetFile.min_offset,
                    )
                )
            )
            .scalars()
            .all()
        )

        out: dict[tuple[str, int], list[ParquetFile]] = {}
        if not rows:
            return out

        now = datetime.datetime.now(datetime.UTC)
        tg = self.config.PARQUET_COMPACTION_TARGET_BYTES
        min_files = self.config.PARQUET_COMPACTION_MIN_INPUT_FILES
        max_files = self.config.PARQUET_COMPACTION_MAX_INPUT_FILES
        force_age_sec = self.config.PARQUET_COMPACTION_FORCE_AGE_SEC

        # group by (topic, partition)
        cur_key = None
        bucket: list[ParquetFile] = []
        for pf in rows:
            key = (pf.topic_name, pf.partition_number)
            if cur_key is None:
                cur_key = key
            if key != cur_key:
                self._maybe_pick_bucket(
                    out, cur_key, bucket, now, tg, min_files, max_files, force_age_sec
                )
                cur_key, bucket = key, []
            bucket.append(pf)
        if bucket:
            self._maybe_pick_bucket(
                out, cur_key, bucket, now, tg, min_files, max_files, force_age_sec
            )

        return out

    @staticmethod
    def _maybe_pick_bucket(
            out, key, files, now, target_bytes, min_files, max_files, force_age_sec
    ):
        if not files:
            return
        # size trigger
        total = 0
        chosen: list[ParquetFile] = []
        for f in files:
            if len(chosen) >= max_files:
                break
            chosen.append(f)
            total += f.total_bytes
            if total >= target_bytes and len(chosen) >= min_files:
                out[key] = chosen
                return
        # age trigger
        oldest = files[0]
        oldest_ts = oldest.created_at or oldest.updated_at
        if (
                oldest_ts
                and (now - oldest_ts).total_seconds() >= force_age_sec
                and len(files) >= min_files
        ):
            out[key] = files[:max_files]

    def _wal_uri_to_object_key(self, uri: str) -> str:
        return wal_uri_to_object_key(self.config, uri)


# currently only support Avro for schema
# in theory can support plain json with some caveats
# also non schema data, ie pure key/value straight from kafka records as key and value columns
class IcebergCompactor(CompactionProcessor):
    async def apply(self, ctx: CompactionContext):
        batches_by_topic: dict[str, list[WALBatch]] = defaultdict(list)
        for wal_file in ctx.wal_decoded:
            for batch in wal_file.batches:
                batches_by_topic[batch.topic].append(batch)
        async with httpx.AsyncClient() as http_client:
            async with ctx.config.async_session_factory() as session:
                topics = await session.execute(select(Topic))
                for topic in topics.scalars():
                    if not topic.schema:
                        log.info(f"no schema for topic {topic.name}")
                        continue


def build_uri(config, key: str) -> str:
    bucket_prefix = (
        f"/{config.WAL_BUCKET_PREFIX.strip('/')}" if config.WAL_BUCKET_PREFIX else ""
    )
    return f"{bucket_prefix}/{key}"




from kio.schema.share_fetch.v0.request import (
    ShareFetchRequest as ShareFetchRequestV0,
)
from kio.schema.share_fetch.v0.request import (
    RequestHeader as ShareFetchRequestHeaderV0,
)
from kio.schema.share_fetch.v0.response import (
    ShareFetchResponse as ShareFetchResponseV0,
)
from kio.schema.share_fetch.v0.response import (
    ResponseHeader as ShareFetchResponseHeaderV0,
)

ShareFetchRequestHeader = ShareFetchRequestHeaderV0

ShareFetchResponseHeader = ShareFetchResponseHeaderV0

ShareFetchRequest = ShareFetchRequestV0

ShareFetchResponse = ShareFetchResponseV0
